from .channel import Channel, ConnectedChannel
from .httpclient import HTTPClient, WebResponse
from .log import Log
from .settings import Settings
from .errors import MissingSetting, ConfigFileNotFound, HTTPDriverException

from time import sleep
from datetime import datetime
from copy import deepcopy


api_users = "https://api.twitch.tv/helix/users?"

class Bridge():
    """description of class"""
    channels: list
    active: bool
    connected: bool
    


    def __init__(self, client, oauth_token=None):
        self.client = client
        self.active = True
        if oauth_token is not None:
            self.web = HTTPClient("Bridge", twitch_oauth=oauth_token)
        else:
            try:
                settings = Settings()
                if settings.has_key("database.user") and settings.has_key("database.pass") and settings.has_key("database.name"):
                    self.web = HTTPClient("Bridge", use_model=True)
                elif settings.has_key("twitch.oauth"):
                    self.web = HTTPClient("Bridge", twitch_oauth=settings.get_setting("twitch.oauth"))
                else:
                    raise MissingSetting("You must provide either 'twitch.oauth' or 'database.user' and 'database.pass' in your settings file.")
            except ConfigFileNotFound:
                raise HTTPDriverException("Twitch WebDriver could not be configured, please provide an 'oauth_token'")
        self.log = Log("CORE")
        self.lastCheck = datetime.now()
        self.__interval = 7
        self.connected = False
        pass


    
    def __init_channels(self, connected: bool):
        self.channels = []
        chns = []
        for chn in self.__get_channel_shortlist():
            response: WebResponse = self.web.GetFromTwitch(api_users+chn)
            if response.code == 200 or response.code == "200":
                for user_data in response.json["data"]:
                    chns.append({
                        "login": user_data["login"],
                        "id": user_data["id"]
                    })
            else:
                self.log.error(f"There was an issue communicating with Twitch. Details: {response.message}")
        for chan in chns:
            if connected:
                channel = ConnectedChannel(self, self.client.nick, self.client.token, chan["login"], chan["id"])
            else:
                channel = Channel(self, self.client.nick, self.client.token, chan["login"], chan["id"])
            self.channels.append(channel)
            

    def __get_channel_shortlist(self):
        out = ""
        counter = 0
        maxCount = 80
        for chn in self.client.channels:
            if counter < maxCount:
                out += f"login={chn}&"
                counter += 1
            else:
                yield out[:-1]
                out = f"login={chn}&"
                counter = 0
        yield out[:-1]


    def __get_channel(self, name: str):
        for channel in self.channels:
            if channel.name == name:
                return channel
        



    def run(self, connected = False):
        self.connected = connected
        self.__init_channels(connected)
        self.lastCheck = datetime.now()
        try:
            for channel in self.channels:
                sleep(2)
                channel.start()
            while self.active:
                sleep(0.0001)
                if (datetime.now() - self.lastCheck).seconds > self.__interval:
                    self.client.check_messages(self.channels)
                    self.heartbeat()
                    self.lastCheck = datetime.now()

        except KeyboardInterrupt:         
            for channel in self.channels:
                channel.close()

    def heartbeat(self):
        for channel in self.channels:
            healthy_thread = channel.on_heartbeat()
            if not healthy_thread:
                name = deepcopy(channel.channel_name)
                self.log.info(f"Rebooting Channel Thread for '{name}'")
                self.remove_channel(name)
                self.log.info(f"Channel Thread '{name}' Closed.")
                self.add_channel(name, connected=self.connected)
                self.log.info(f"Channel Thread '{name}' Reopened.")


    def send_message_to_channel(self, channel: str, message: str):
        for chn in self.channels:
            if chn.channel_name == channel:
                chn.send(message)

    def set_interval(self, interval: int):
        self.__interval = interval
    
    
    def add_channel(self, channel: str, id: int = None, connected: bool = False):
        if id is None:
            response: WebResponse = self.web.GetFromTwitch(f"{api_users}login={channel}")
            if response.code == 200 or response.code == "200":
                id = response.json["data"][0]["id"]
            else:
                self.log.error(f"Could not Add Channel. There was an issue communicating with Twitch. Details: {response.message}")
                return
        if connected or self.connected:
            chn = ConnectedChannel(self, self.client.nick, self.client.token, channel, id=id)
        else:
            chn = Channel(self, self.client.nick, self.client.token, channel, id=id)
        self.channels.append(chn)
        chn.start()
        chn.on_new_connection(channel)
        self.client.on_add_channel(channel, id)
    
    def remove_channel(self, channel: str):
        for chn in self.channels:
            if chn.channel_name == channel:
                chn.close()
                self.channels.remove(chn)
        self.client.on_remove_channel(channel)





    





from .errors import MissingSetting, ConfigFileNotFound
from pathlib import Path
import yaml
import json

import threading

lock = threading.Lock()

class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



class Settings(metaclass=Singleton):
    __settingsData: dict
    
    def __init__(self, path=None):
        if path is None:
            path = "config.yaml"
        if not Path(path).exists():
            path = "./config.yaml"
        if not Path(path).exists():
            path = "./bots/config.yaml"
        if not Path(path).exists():
            path = "./Common/config.yaml"
        if not Path(path).exists():
            path = "../Common/config.yaml"
        if not Path(path).exists():
            path = "../../Common/config.yaml"
        if path is None:
            path = "config.json"
        if not Path(path).exists():
            path = "./config.json"
        if not Path(path).exists():
            path = "./bots/config.json"
        if not Path(path).exists():
            path = "./Common/config.json"
        if not Path(path).exists():
            path = "../Common/config.json"
        if not Path(path).exists():
            path = "../../Common/config.json"
        if not Path(path).exists():
            raise ConfigFileNotFound("Config file could not be found.")
        self.__settingsData = self.__parse_file(path)
        self.path = path
    
    def get_setting(self, key:str):
        try:
            setting = self.__fetchSetting(key, self.__settingsData)
        except KeyError:
            raise MissingSetting(f"Key: '{key}' was not found in the config file '{self.path}'")
        if setting is None:
            raise MissingSetting(f"Key: '{key}' was not found in the config file '{self.path}'")
        return self.__fetchSetting(key, self.__settingsData)

    def has_key(self, key:str):
        try:
            self.get_setting(key)
            return True
        except MissingSetting:
            return False
    
    def __fetchSetting(self, key:str, sublist:dict):
        keynest = key.split(".", 1)
        if len(keynest) == 1:
            try:
                return sublist[keynest[0]]
            except:
                return None
        return self.__fetchSetting(keynest[1], sublist[keynest[0]])
        
    def __parse_file(self, path):
        data = None
        with open(Path(path), 'r') as stream:
            if path.endswith(".yaml"):
                data = yaml.safe_load(stream)
            elif path.endswith(".json"):
                data = json.load(stream)
            else:
                raise ConfigFileNotFound("Config file must be a .yaml or a .json file")
        return data
    
    def get_all_settings(self):
        return self.__settingsData

    def __repr__(self):
        return f"[Settings]: Loaded {len(self.__settingsData)} settings from file: '{self.path}'"


from requests import get, patch, put, post
from .settings import Settings
from .log import Log
from .model import Model
from .errors import DBConnectionUndefined, HTTPDriverNotConfigured
from urllib import parse
from json.decoder import JSONDecodeError

class WebResponse():
    def __init__(self, data:dict):
        self.status = data["status"]
        self.message = data["message"]
        self.code = data["code"]
        self.text = data["text"]
        self.json = data["json"]
        self.headers = data["headers"]

    def __repr__(self):
        return f"[{self.status}]({self.message}): {self.text}"


class HTTPClient():

    def __init__(self, owner:str, use_model=False, twitch_oauth=None):
        self.log = Log("Web Handler", f"Web Handler: {owner}")
        self.useModel = False
        if use_model:
            self.model = Model()
            self.useModel = True
        self.twitchToken = twitch_oauth

    def GetFromTwitch(self, url: str) -> WebResponse:
        headers = self.__get_twitch_headers(url)
        return self.Get(url, headers=headers)

    def GetFromDiscord(self, url: str) -> WebResponse:        
        headers = self.__get_discord_headers()
        return self.Get(url, headers=headers)

    def PutToTwitch(self, url: str, body: dict) -> WebResponse:
        headers = self.__get_twitch_headers(url)
        return self.Put(url, body=body, headers=headers)
    
    def PutJSONToTwitch(self, url: str, body: dict) -> WebResponse:
        headers = self.__get_twitch_headers(url)
        return self.PutJSON(url, body=body, headers=headers)

    def PostFormToTwitch(self, url:str,  body: dict) -> WebResponse:
        headers = self.__get_twitch_headers(url)
        return self.PostForm(url=url, body=body, headers=headers)

    def PostFormToDiscord(self, url:str,  body: dict) -> WebResponse:
        headers = self.__get_discord_headers()
        return self.PostForm(url=url, body=body, headers=headers)

    def PostJSONToTwitch(self, url:str,  body: dict) -> WebResponse:
        headers = self.__get_twitch_headers(url)
        return self.PostJSON(url=url, body=body, headers=headers)

    def PostJSONToDiscord(self, url:str,  body: dict) -> WebResponse:
        headers = self.__get_discord_headers()
        return self.PostJSON(url=url, body=body, headers=headers)



    def Get(self, url: str, headers: dict = None) -> dict:
        """Get request"""           
        self.log.debug(f"Sending new GET request to '{url}'.\nHeaders: \n{headers}")
        try:
            response = get(url=url, headers=headers)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt

    def Put(self, url: str, body: dict, headers: dict = None):
        """Put request"""           
        self.log.debug(f"Sending new PUT request to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        try:
            response = put(url=url, data=body, headers=headers)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt
    
    def PutJSON(self, url: str, body: dict, headers: dict = None):
        """Put request"""           
        self.log.debug(f"Sending new PUT request to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        try:
            response = put(url=url, json=body, headers=headers)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt


    
    def PostForm(self, url:str,  body: dict, headers: dict = None):
        if headers is None:
            setting = Settings()
            headers = {
                "Client-ID": setting.get_setting("tokens.twitch.user"), 
                "content-type": 'application/x-www-form-urlencoded', 
            }
        self.log.debug(f"Sending new POST request to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        try:
            response = post(url=url, headers=headers, data=body)
        except Exception as ex:
            return self.__exception_response(str(ex))
        rt = self.__process_response(response)
        self.log.debug("returning: \n{}".format(rt))
        return rt
    
    def PostJSON(self, url:str,  body: dict, headers: dict = None):
        if headers is None:
            setting = Settings()
            headers = {
                "Client-ID": setting.get_setting("tokens.twitch.user"), 
                "content-type": 'application/json', 
            }
        self.log.debug(f"Sending new POST request with JSON body to '{url}'.\nHeaders: \n{headers}\nBody:\n{body}")
        response = post(url=url, headers=headers, json=body)
        try:
            rt = self.__process_response(response)
        except Exception as ex:
            return self.__exception_response(str(ex))
        self.log.debug("returning: \n{}".format(rt))
        return rt
    
    def __exception_response(self, message):
        self.log.warn("An exception occrred when attempting to contact the server. Exception: " + message)
        return WebResponse ({
            "status": "error",
            "message": "an exception occurred when attempting to contact server.",
            "code": 600,
            "text": message,
            "json": {"error": message},
            "headers": None
        })


    def __process_response(self, response):
        if response.status_code == 200:
            self.log.debug("Response: 200 | OK \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "response ok",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 201:
            self.log.debug("Response: 201 | OK \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "no data returned",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 202:
            self.log.debug("Response: 202 | Accepted \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "accepted",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 204:
            self.log.debug("Response: 204 | No Content \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "ok",
                "message": "no content",
                "code": response.status_code,
                "text": "No Data",
                "json": {"No Data"},
                "headers": response.headers
            })
        elif response.status_code == 400:
            self.log.info("Response: 400 | Bad Request \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "bad request",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 401:
            self.log.info("Response: 401 | Unauthorised Request \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "unauthorised request",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        elif response.status_code == 404:
            self.log.info("Response: 404 | Resource not found \n Headers:\n{}\n\nContent:\n{}".format(response.headers, response.text))
            return WebResponse ({
                "status": "error",
                "message": "resource not found",
                "code": response.status_code,
                "text": response.text,
                "json": response.json(),
                "headers": response.headers
            })
        else:
            self.log.error("Response: {} | \nHeaders: {} \nError Response\n{}".format(response.status_code, response.headers, response.text))
            try:
                return WebResponse ({               
                    "status": "error",
                    "message": "server returned an error",
                    "code": response.status_code,
                    "text": response.text,
                    "json": response.json(),
                    "headers": response.headers
                })
            except JSONDecodeError:
                return WebResponse ({               
                    "status": "error",
                    "message": "server returned an error",
                    "code": response.status_code,
                    "text": response.text,
                    "json": {"JSON not available"},
                    "headers": response.headers
                })

    def __get_twitch_headers(self, url: str) -> dict:
        setting = Settings()
        token = ""

        if self.twitchToken is None and not self.useModel:
            raise HTTPDriverNotConfigured("Values provided for DB connection / Twitch OAuth are invalid.\nYou must provide either:\n - Values for the Databse with 'db_user' and 'db_pass'\n - An OAuth token for the Twitch API")

        elif self.twitchToken is not None:
            token = self.twitchToken
        
        else:
            jOAUTH = self.model.fetchOne("select AccessToken from Tokens where TokenType = 'JarvisAppToken'")
            token = jOAUTH[0]

        if "kraken" in url:
            headers = {
                "Client-ID" : setting.get_setting("tokens.twitch.user"),
                "Authorization": "OAuth " + token,
                "Accept": "application/vnd.twitchtv.v5+json"
            }
        else:
            headers = {
                "Client-ID" : setting.get_setting("tokens.twitch.user"),
                "Authorization": "Bearer " + token
            }

        return headers

    
    def __get_discord_headers(self) -> dict:
        setting = Settings()
        headers = {
            "Client-ID": "{}".format(setting.get_setting("integration.discord.client")),
            "User-Agent": setting.get_setting("integration.discord.agent"),
            "Authorization": "Bot {}".format(setting.get_setting("integration.discord.key"))
        }
        return headers


from requests import post
from time import time

from .settings import Settings
from .errors import TelemetryUnavailable
from .log import Log




class Telemetry():
    sender: str
    

    def __init__(self, sender: str):
        self.log = Log(f"Telemetry: {sender}", "Telemetry", "log")
        self.sender = sender

    def low_priority(self,  message: str, urlTitle: str=None, url: str=None):
        self.__send_message(-1, message, urlTitle=urlTitle, url=url)

    def normal_priority(self,  message: str, urlTitle: str=None, url: str=None):
        self.__send_message(0, message, urlTitle=urlTitle, url=url)

    def high_priority(self,  message: str, urlTitle: str=None, url: str=None):
        self.__send_message(1, message, urlTitle=urlTitle, url=url)

    def critical_priority(self,  message: str, urlTitle: str=None, url: str=None):
        self.__send_message(2, message, urlTitle=urlTitle, url=url)
        


    def __send_message(self, priority: int, message: str, urlTitle: str=None, url: str=None):
        settings = Settings()
        
        if not settings.has_key("tokens.pushover.user"):
            raise TelemetryUnavailable("The pushover settings have not been set in the setting file. These must be set first before attempting to use Telemetry.")

        self.log.log(f"Telemetry Sent:\n| Message | {message} |\n|Url|[{urlTitle}]({url})|")
        
        body = {
            "token": settings.get_setting("tokens.pushover.token"),
            "user": settings.get_setting("tokens.pushover.user"),
            "html": 1,
            "title": self.sender,
            "message": message,
            "timestamp": time(),
            "priority": priority
        }

        if urlTitle is not None and url is not None:
            body["url_title"] = urlTitle
            body["url"] = url
        elif url is not None: 
            body["url"] = url

        response = post(
            "https://api.pushover.net/1/messages.json",
            json=body
        )

        self.log.log(f"Telemetry received the response from pushover:\n| Status Code | {response.status_code} |\n| Response Text | {response.text} |")
